import aiohttp
import json
from typing import Any, Optional, Mapping
from .status import Status
from .pattern import PatternMeta, PatternBuilder


class Flickerstrip:
    def __init__(self, host: str, port: int = 80):
        """The initializer for the Flickerstrip class.

        Args:
            host (string): The host of the flickerstrip.
            port (int, optional): The port of the flickerstrip.
            Defaults to port 80.
        """
        self.host: str = host
        self.port: int = port
        self.status: Status | None = None

    async def __handle_response(self, response):
        """Check if the request was succesful.

        Args:
            response (requests.Response): The response from the request.

        Returns:
            bool: If the request was successful
        """
        body = await response.text()

        if body == "Not Found":
            print(f"ERROR: Request failed with message \"{body}\".")
            return False

        data = json.loads(body)

        if "type" not in data:
            if "id" in data:  # create pattern response.
                return True
            print("ERROR: Unable to get response type!")
            print(data)
            return False

        respType = data["type"]
        match respType:
            case "error":
                message = data["message"]
                print(f"ERROR: Request failed with message \"{message}\".")
                return False
            case "OK":
                return True
            case "status":
                self.status = Status.from_json(data)
                return True
            case _:
                print(f"ERROR: Unrecognized response type: {respType}.")
                print(data)
                return False

    async def __get_request(self, path: str,
                            params: Optional[Mapping[str, str]] = None,
                            return_body: bool = False,
                            ) -> bool | bytes:
        """Send a GET request to the strip.

        Args:
            path (string): The path for the URL to request.
            params (dictionary, optional): The query string params for the
            request. Defaults to None.

        Returns:
            ClientResponse: the client response
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(f"http://{self.host}:{self.port}/{path}",
                                   params=params) as response:
                if return_body:
                    return await response.read()
                else:
                    return await self.__handle_response(response)

    async def __post_request(self, path,
                             params: Optional[Mapping[str, str]] = None,
                             json: Any = None,
                             data: Any = None
                             ) -> bool:
        """Send a POST request to the strip.

        Args:
            path (string): The path for the URL to request.
            json (dictionary, optional): The JSON encodable body for the
            request. Defaults to None.
            params (dictionary, optional): The query string params for the
            request. Defaults to None.
            data (string, optional): The raw data body. Defaults to None.

        Returns:
            ClientResponse: the client response
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(f"http://{self.host}:{self.port}/{path}",
                                    params=params, json=json, data=data
                                    ) as response:
                return await self.__handle_response(response)

    async def force_update(self) -> bool:
        """Force updates the status of the strip.

        Returns:
            bool: If the request was succesful
        """
        return bool(await self.__get_request("status"))

    async def power_on(self) -> bool:
        """Trigger the strip to turn the lights on.

        Returns:
            bool: If the request was succesful
        """
        return bool(await self.__get_request("power/on"))

    async def power_off(self) -> bool:
        """Trigger the strip to turn the lights off.

        Returns:
            bool: If the request was succesful
        """
        return bool(await self.__get_request("power/off"))

    async def power_toggle(self) -> bool:
        """Trigger the strip to toggle the power status.

        Returns:
            bool: If the request was succesful
        """
        return bool(await self.__get_request("power/toggle"))

    async def next_pattern(self) -> bool:
        """Trigger the strip to switch to the next pattern in memory.

        Returns:
            bool: If the request was succesful
        """
        return bool(await self.__get_request("pattern/next"))

    async def freeze_frame(self, frame: int) -> bool:
        """Freeze the animation at the specified frame.

        Args:
            frame (int): The frame number to freeze on.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(frame)}
        return bool(await self.__get_request("pattern/frame", params=params))

    async def download_pattern(self, id: int) -> PatternBuilder | None:
        """Download a pattern by its id

        Args:
            id (int): The ID of the pattern

        Returns:
            PatternBuilder: a pattern builder with the data downloaded
        """

        if self.status is None:
            return None
        meta = self.status.get_pattern_by_id(id)
        if meta is None:
            return None

        params = {"id": str(id)}
        data = bytes(await self.__get_request("pattern/download",
                                              params=params, return_body=True))

        return PatternBuilder.from_binary_string(data, meta)

    async def select_pattern(self, index: int) -> bool:
        """Change the current pattern to the pattern at the specified index.

        Args:
            index (int): The index of the pattern.

        Returns:
            bool: If the request was succesful
        """
        params = {"index": str(index)}
        return bool(await self.__get_request("pattern/select", params=params))

    async def delete_pattern(self,
                             pattern: Optional[PatternMeta] = None,
                             index: Optional[int] = None,
                             id: Optional[int] = None) -> bool:
        """Delete a pattern by its meta, index, or id

        Args:
            pattern (PatternMeta, optional): The pattern meta.
            index (int, optional): The index of the pattern.
            id (int, optional): The id of the pattern.

        Returns:
            bool: If the request was succesful
        """
        if pattern is not None:
            if not isinstance(pattern, PatternMeta):
                raise TypeError(
                    "Expected a PatternMeta object for the pattern arg.")
            return await self.delete_pattern(id=pattern.id)
        elif index is not None:
            params = {"index": str(index)}
            return bool(await self.__get_request("pattern/forget",
                                                 params=params))
        elif id is not None:
            params = {"id": str(id)}
            return bool(await self.__get_request("pattern/forget",
                                                 params=params))
        else:
            raise TypeError(
                "Deleting a pattern requires one of the three args.")

    async def create_pattern(self,
                             pattern: PatternBuilder,
                             preview=True
                             ) -> bool:
        """Create a pattern on the flickerstrip.

        Args:
            pattern (PatternBuilder): The pattern to send.
            preview (bool, optional): Whether to only preview the pattern.
            Defaults to True.

        Returns:
            bool: If the request was succesful
        """
        params = {
            "name": pattern.name,
            "fps": str(pattern.fps),
            "frames": str(pattern.frame_count),
            "pixels": str(pattern.pixel_count),
            "preview": str(1 if preview else 0)
        }
        data = pattern.to_binary_string()
        return await self.__post_request("pattern/create",
                                         params=params, data=data)

    async def set_color(self, r: int, g: int, b: int) -> bool:
        """Sets the strip to a solid color (given as an RGB value).

        Args:
            r (int): The red value (0 - 255)
            g (int): The green value (0 - 255)
            b (int): The blue value (0 - 255)

        Returns:
            bool: If the request was successful
        """
        pattern = PatternBuilder("Solid")
        pattern.add_pixel(r, g, b)
        return await self.create_pattern(pattern)

    async def set_brightness(self, value: int) -> bool:
        """Set the brightness of the flickerstrip.

        Args:
            value (int): The brightness level as an int from 0 to 100.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("brightness", params=params))

    async def set_name(self, value: str) -> bool:
        """Set the name of the flickerstrip.

        Args:
            value (string): The new name.

        Returns:
            bool: If the request was succesful
        """
        json = {"name": value}
        return bool(await self.__post_request("config/name", json=json))

    async def set_group(self, value: str) -> bool:
        """Set the group of the flickerstrip.

        Args:
            value (string): The new group.

        Returns:
            bool: If the request was succesful
        """
        json = {"name": value}
        return bool(await self.__post_request("config/group", json=json))

    async def set_cycle(self, value: int) -> bool:
        """Set the cycle timer of the flickerstrip.

        Args:
            value (int): How long to cycle to next pattern.
            Value is handled in seconds.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("config/cycle", params=params))

    async def set_fade_duration(self, value: int) -> bool:
        """Set the fade timer of the flickerstrip.

        Args:
            value (int): How long to fade to next pattern.
            Value is handled in seconds.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("config/fade", params=params))

    async def set_strip_length(self, value: int) -> bool:
        """Set the length of the flickerstrip.

        Args:
            value (int): The length of the strip in pixels.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("config/length", params=params))

    async def set_strip_start(self, value: int) -> bool:
        """Set the start pixel of the strip.

        This will make the flickerstrip ignore all pixels before
        the given number, set to 0 for the first pixel.

        Args:
            value (int): The new first pixel of the strip.

        Returns:
            bool: If the request was successful.
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("config/start", params=params))

    async def set_strip_end(self, value: int) -> bool:
        """Set the end pixel of the strip.

        This will make the flickerstrip ignore all pixels after
        the given number, set to -1 for the last pixel.

        Args:
            value (int): The new last pixel of the strip.

        Returns:
            bool: If the request was successful.
        """
        params = {"value": str(value)}
        return bool(await self.__get_request("config/end", params=params))

    async def set_reversed(self, value: bool) -> bool:
        """Set the reversed state of the flickerstrip.

        Args:
            value (bool): If the flickerstrip should animate
            patterns in reverse.

        Returns:
            bool: If the request was succesful
        """
        params = {"value": str(1 if value else 0)}
        return bool(await self.__get_request("config/reversed", params=params))
